local Maths  = require 'lib/maths'

local Pause    = require 'src/states/pause'
local GameOver = require 'src/states/game_over' 

local Coin    = require 'src/objects/coin'
local PowerUp = require 'src/objects/power_up' 
local Laser   = require 'src/objects/laser'

local level, coins, power_ups, player, laser_cannons, lasers, ui
local max_abilities = 5

local speed_increaser = 5
local max_speed = 65
local speed_ticker = 30
local speed_tick = 0
local increase_speed = love.audio.newSource('sfx/increase_speed.wav', 'static')

local debris_img = love.graphics.newImage('img/debris.png')
debris_img:setFilter('nearest', 'nearest')
local debris_quads = {
  love.graphics.newQuad( 0, 0, 8, 8, debris_img:getWidth(), debris_img:getHeight()),
  love.graphics.newQuad( 8, 0, 8, 8, debris_img:getWidth(), debris_img:getHeight()),
  love.graphics.newQuad(16, 0, 8, 8, debris_img:getWidth(), debris_img:getHeight()),
  love.graphics.newQuad(32, 0, 8, 8, debris_img:getWidth(), debris_img:getHeight()),
  love.graphics.newQuad(40, 0, 8, 8, debris_img:getWidth(), debris_img:getHeight()),
}
local debris = love.graphics.newParticleSystem(debris_img, 8)
local debris_coin = love.audio.newSource('sfx/coin2.ogg', 'static')

state_play = newState()

function state_play.load(vars)
  level     = vars.level
  coins     = {}
  power_ups = {}
  player    = vars.player
  ui        = vars.ui

  debris:setBufferSize(8)
  debris:setParticleLifetime(0.4)
  debris:setSizes(1.5, 0.2)
  debris:setSpeed(100)
  debris:setLinearAcceleration(0, 300)
  debris:setDirection(math.rad(-90))
  debris:setSpread(math.rad(45))
  debris:setRelativeRotation(true)
  debris:setEmissionArea('none')
  debris:setQuads(debris_quads[1])
  
  camera.setPosition(0, 0)
  camera.speed = 10
  speed_tick = -10
  background.speed = camera.speed / 2

  laser_cannons = vars.laser_cannons
  for _, v in pairs(laser_cannons) do v.setAnimation('roll') end

  lasers = {}
  table.insert(lasers, newLaser(laser_cannons[1].x + (level.tile_size/2), laser_cannons[1].y - (level.tile_size/2)))
  table.insert(lasers, newLaser(laser_cannons[3].x + (level.tile_size/2), laser_cannons[3].y - (level.tile_size/2)))

  music:stop()
  music:play()
end

function state_play.update(dt)
  background.update(dt)
  -- Camera Functions
  speed_tick = speed_tick + dt
  if speed_tick >= speed_ticker then
    speed_tick = speed_tick - speed_ticker
    camera.speed = camera.speed >= max_speed and max_speed or camera.speed + speed_increaser
    background.speed = camera.speed / 2
    increase_speed:play()
  end
  camera.move(0, camera.speed * dt)
  if camera.y >= level.tile_size then
    local spawnx, spawny = 0, 0
    spawnx, spawny = level.spawnFloor()
    if spawnx > 0 then
      -- Spawn Power-up: 1 in 6 Chances
      if love.math.random(6) == 6 then
        table.insert(power_ups, newPowerUp(spawnx, spawny))
      else
        table.insert(coins, newCoin(spawnx, spawny, camera.speed))
      end
    end
    ui.score = ui.score + math.ceil(0.25 * camera.speed)

    camera.move(0, -level.tile_size)
    for _, coin in pairs(coins) do coin.move(0, -level.tile_size) end
    for _, pw in pairs(power_ups) do pw.move(0, -level.tile_size) end
    player.move(0, -level.tile_size)
  end

  -- For Ability & Collision Detection
  local pleft  = player.x - player.radius
  local pright = player.x + player.radius

  local l, x, y = level.tiles.getTile(math.ceil(pleft/level.tile_size),     math.ceil(player.y/level.tile_size)+1)
  local l2, x2, y2 = level.tiles.getTile(math.ceil(pright/level.tile_size), math.ceil(player.y/level.tile_size)+1)

  -- Player Ability Input
  if love.keyboard.isDown('down') and ui.cooldown <= 0 and player.state == 'walking' and ui.jackhammers > 0 then
    -- Use up a Jack Hammer
    ui.jackhammers = ui.jackhammers - 1
    ui.cooldown = 2
    
    -- Remove Tiles Below Player
    level.drillTiles(math.ceil(pleft/level.tile_size), math.ceil(player.y/level.tile_size)+1)
    
    -- Random Treasure Found in Drilled Area
    local debris_count = love.math.random(debris:getBufferSize())
    local adder = 0
    if love.math.random(5) == 5 then
      adder = math.ceil(camera.speed / 25) * debris_count
      debris:setQuads(debris_quads[2])
    elseif camera.speed >= 20 and love.math.random(7) == 7 then
      adder = math.ceil(camera.speed / 20) * debris_count
      debris:setQuads(debris_quads[3])
    elseif camera.speed >= 40 and love.math.random(10) == 10 then
      adder = math.ceil(camera.speed / 15) * debris_count
      debris:setQuads(debris_quads[4])
    elseif camera.speed >= 60 and love.math.random(80) == 80 then
      adder = math.ceil(camera.speed / 10) * debris_count
      debris:setQuads(debris_quads[5])
    else
      debris:setQuads(debris_quads[1])
    end
    
    if adder > 0 then
      debris_coin:play()
      ui.addMessage(adder, player.x, player.y)
    end
    
    -- Debris flies over the drilled area
    debris:moveTo(player.x, player.y)
    debris:emit(debris_count)
  end
  
  -- Player Movement Inputs
  player.inputs(dt)

  -- Player: Coin Collision
  for _, coin in pairs(coins) do
    if getCircleCircleDistance(coin.x, coin.y, coin.radius, player.x, player.y, player.radius) < 0 and coin.obtained == false then
      coin.obtained = true
      coin.sound:play()
      ui.addMessage(coin.score, coin.x, coin.y)
    end
  end
  
  -- Player: Power-ups Collision
  for _, pw in pairs(power_ups) do
    if getCircleCircleDistance(pw.x, pw.y, pw.radius, player.x, player.y, player.radius) < 0 and pw.obtained == false then
      pw.obtained = true
      pw.sound:play()
      ui.addJackHammer()
      local adder = math.ceil(camera.speed/20)
      ui.addMessage(adder, pw.x, pw.y)
    end
  end

  -- Delete Coins
  local deleter = {}
  for i = 1, #coins do
    if coins[i].y + coins[i].radius < 0 or coins[i].obtained then
      table.insert(deleter, i)
    end
  end
  for i = #deleter, 1, -1 do
    table.remove(coins, deleter[i])
  end

  -- Delete Power Ups  
  deleter = {}
  for i = 1, #power_ups do
    if power_ups[i].y + power_ups[i].radius < 0 or power_ups[i].obtained then
      table.insert(deleter, i)
    end
  end
  for i = #deleter, 1, -1 do
    table.remove(power_ups, deleter[i])
  end

  -- Player: Falling Collision
  collide = (l > 0 and pleft + 2 < x + level.tile_size) or (l2 > 0 and pright - 2 > x2)
  if player.state == 'walking' and not collide then
    player.state = 'falling'
  elseif player.state == 'falling' and collide then
    player.state = 'walking'
    player.y = y - player.radius
  elseif player.vy > 0 then
    player.state = 'falling'
  end

  -- Player: Jump Collision
  l, x, y = level.tiles.getTile(math.ceil(pleft/level.tile_size), math.floor(player.y/level.tile_size))
  l2, x2, y2 = level.tiles.getTile(math.ceil(pright/level.tile_size), math.floor(player.y/level.tile_size))
  local collide = (l > 0 and pleft + 2 < x + level.tile_size) or (l2 > 0 and pright - 2 > x2)
  if player.state == 'jumping' and collide then
    player.y = y + level.tile_size + player.radius
  end

  -- Player: MovingLeft Collision
  if player.vx < 0 then
    l, x, y = level.tiles.getTile(math.ceil(pleft/level.tile_size), math.ceil(player.y/level.tile_size))
    if l > 0 and pleft < x + level.tile_size then
      player.x = x + level.tile_size + player.radius + 1
    end
  end
  -- Player: MovingRight Collision
  if player.vx > 0 then
    l, x, y = level.tiles.getTile(math.ceil((pright)/level.tile_size), math.ceil(player.y/level.tile_size))
    if l > 0 and pright > x then
      player.x = x - player.radius - 1
    end
  end

  -- Player: Laser Collision [[ GAME OVER DETECTOR ]]
  lasers[1].hit = player.y - player.radius < lasers[1].y + 12 + camera.y
  lasers[2].hit = player.y + player.radius > lasers[2].y +  2 + camera.y
  if lasers[1].hit or lasers[2].hit then
    -- Game Over
    music:stop()
    debris:setPosition(player.x, player.y)
    debris:setBufferSize(30)
    debris:setSpeed(50)
    debris:setParticleLifetime(1)
    debris:setLinearAcceleration(-100,200,100)
    debris:setEmissionArea('uniform', 5, 5, 0)
    debris:setSizes(0.2)
    debris:setQuads(debris_quads[1])
    debris:emit(30)
    state_machine.change(state_gameover, {level = level, debris = debris, laser_cannons = laser_cannons, score = ui.score})
  end

  player.update(dt)
  debris:update(dt)
  lasers[1].update(dt)
  lasers[2].update(dt)
  ui.update(dt)

  -- Animations
  for _, coin in pairs(coins) do coin.updateAnimation(dt) end
  for _, pw in pairs(power_ups) do pw.updateAnimation(dt) end
  player.updateAnimation(dt)
  for _, lc in pairs(laser_cannons) do lc.updateAnimation(dt) end
  
end
function state_play.draw()
  background.draw()
  camera.set()
    level.draw()
    for _, coin in pairs(coins) do coin.draw() end
    for _, pw in pairs(power_ups) do pw.draw() end
    player.draw()
  camera.unset()

  for _, lc in pairs(laser_cannons) do lc.draw() end

  lasers[1].draw()
  lasers[2].draw()
  
  love.graphics.draw(debris)
  
  ui.draw()
end

function state_play.keypressed(key)
  if key == 'escape' then
    music:pause()
    state_machine.push(state_pause)
  end
end
